import time
from pathlib import Path
from typing import Dict, Any, Optional, Protocol, List, Union
from frappe_manager.compose_project.compose_project import ComposeProject
from frappe_manager.docker_wrapper.DockerException import DockerException
from frappe_manager.services_manager.services_exceptions import (
    DatabaseServiceDBCreateFailed,
    DatabaseServiceDBExportFailed,
    DatabaseServiceDBImportFailed,
    DatabaseServiceDBNotFoundError,
    DatabaseServiceDBRemoveFailError,
    DatabaseServiceException,
    DatabaseServicePasswordNotFound,
    DatabaseServiceStartTimeout,
    DatabaseServiceUserRemoveFailError,
)
from frappe_manager.display_manager.DisplayManager import richprint
from pydantic import BaseModel

from frappe_manager.docker_wrapper.subprocess_output import SubprocessOutput


# TODO this class will be used for validation for main config
class DatabaseServerServiceInfo(BaseModel):
    host: str
    user: str
    port: int
    password: str
    secret_path: Path

    @classmethod
    def import_from_compose_file(cls, compose_service_name: str, compose_project: ComposeProject):
        """
        Provides info about a database server
        """
        compose_service_envs = compose_project.compose_file_manager.get_envs(container=compose_service_name)

        info: Dict[str, Any] = {}
        info["user"] = 'root'
        # this also being considered as servicename
        info["host"] = compose_service_name
        info["port"] = 3306

        # TODO use fm main config here
        # secrets or password ?
        if 'MYSQL_ROOT_PASSWORD_FILE' in compose_service_envs:
            password_path: Path = compose_project.compose_file_manager.get_secret_file_path('db_root_password')
            info['secret_path'] = password_path
            info["password"] = password_path.read_text()
        elif 'MYSQL_ROOT_PASSWORD' in compose_service_envs:
            info['secret_path'] = 'env'
            info["password"] = compose_service_envs['MYSQL_ROOT_PASSWORD']
        else:
            raise DatabaseServicePasswordNotFound(compose_service_name)

        return cls(**info)


class DatabaseServiceManager(Protocol):
    database_server_info: DatabaseServerServiceInfo
    compose_project: ComposeProject

    def __init__(self, database_server_info: DatabaseServerServiceInfo, compose_project: ComposeProject) -> None:
        ...

    def remove_user(self, db_user: str, db_user_host: str = '%', remove_all_host: bool = False):
        ...

    def add_user(self, db_user: str, db_pass: str, db_user_host: str = '%', force: bool = False, timeout=25):
        ...

    def grant_user_privilages(self, db_user: str, db_name: str):
        ...

    def check_user_exists(self, db_user: str):
        ...

    def check_db_exists(self, db_name: str):
        ...

    def remove_db(self, db_name: str):
        ...

    def wait_till_db_start(self, interval: int = 5, timeout: int = 30) -> bool:
        ...

    def db_import(self, db_name: str, host_db_file_path: Path, force: bool = False):
        ...


class MariaDBManager(DatabaseServiceManager):
    def __init__(
        self,
        database_server_info: DatabaseServerServiceInfo,
        compose_project: ComposeProject,
        run_on_compose_service: Optional[str] = None,
    ) -> None:
        """
        Database manager
        """
        self.database_server_info: DatabaseServerServiceInfo = database_server_info
        self.compose_project: ComposeProject = compose_project

        if not run_on_compose_service:
            self.run_on_compose_service: str = self.database_server_info.host
        else:
            self.run_on_compose_service: str = run_on_compose_service

        self.base_command = f"/usr/bin/mariadb -u{self.database_server_info.user} -p'{self.database_server_info.password}' -P{self.database_server_info.port} -h{self.database_server_info.host} "
        self.base_query = '-e '
        self.quiet = True

    def db_run_query(
        self, query: str, raise_exception_obj: Optional[DatabaseServiceException] = None, capture_output: bool = False
    ):
        base_command = self.base_command

        if capture_output:
            base_command += '--batch --skip-column-names '

        db_query = base_command + self.base_query + query

        try:
            output = self.compose_project.docker.compose.exec(
                self.run_on_compose_service, command=db_query, stream=not capture_output
            )
            if capture_output:
                return output
            richprint.live_lines(output)
        except DockerException as e:
            if raise_exception_obj:
                raise raise_exception_obj
            raise e

    def wait_till_db_start(self, interval: int = 5, timeout: int = 30) -> bool:
        for i in range(timeout):
            if not self.is_db_running():
                time.sleep(interval)
            else:
                return True
        total_timeout = interval * timeout
        raise DatabaseServiceStartTimeout(total_timeout, self.run_on_compose_service)

    def is_db_running(self) -> bool:
        db_started_command = f"mysqladmin  -P{self.database_server_info.port} -h{self.database_server_info.host} -u'{self.database_server_info.user}' -p'{self.database_server_info.password}' ping"
        try:
            output = self.compose_project.docker.compose.exec(
                self.run_on_compose_service, command=db_started_command, stream=False
            )
            return 'mysqld is alive' in " ".join(output.stdout)
        except DockerException as e:
            return False

    def get_db_users(self) -> Dict[str, str]:
        show_db_user_command = f"'SELECT User, Host FROM mysql.user;'"
        exception = DatabaseServiceException(self.database_server_info.host, 'Failed to determine mysql users.')
        output: SubprocessOutput = self.db_run_query(
            show_db_user_command, raise_exception_obj=exception, capture_output=True
        )
        user_list: Dict[str, str] = {}
        for line in output.stdout:
            username, host = line.split('\t')
            user_list[username] = host
        return user_list

    def check_user_exists(self, username: str, host: Optional[str] = None) -> bool:
        user_list = self.get_db_users()
        if not username in user_list:
            return False
        if not host:
            return True
        if not user_list[username] == host:
            return False
        return True

    def get_all_databases(self) -> List[str]:
        db_exits_commmand = f"'SHOW DATABASES;'"
        db_exits_exception = DatabaseServiceException(
            self.database_server_info.host, 'Failed to get list of all databases.'
        )
        output: SubprocessOutput = self.db_run_query(db_exits_commmand, db_exits_exception, capture_output=True)
        return output.stdout

    def check_db_exists(self, db_name: str):
        databases = self.get_all_databases()
        return db_name in databases

    def remove_user(self, db_user: str, db_user_host: str = '%', remove_all_host: bool = False):
        users = {db_user: db_user_host}

        if remove_all_host:
            users = self.get_db_users()

        for user, host in users.items():
            if db_user == user:
                remove_db_user_command = f"'DROP USER `{user}`@`{host}`;'"
                remove_db_user_exception = DatabaseServiceUserRemoveFailError(user, host)
                self.db_run_query(remove_db_user_command, remove_db_user_exception)

    def remove_db(self, db_name: str):
        remove_db_command = f"'DROP DATABASE `{db_name}`;'"
        remove_db_exception = DatabaseServiceDBRemoveFailError(db_name, self.database_server_info.host)
        self.db_run_query(remove_db_command, remove_db_exception)

    def grant_user_privilages(self, db_user: str, db_name: str):
        grant_user_command = f"'GRANT ALL PRIVILEGES ON `{db_name}`.* TO `{db_user}`@`%`;'"
        grant_user_exception = DatabaseServiceException(
            self.database_server_info.host, f'Failed to grant prvilages for user {db_user} on {db_name}.'
        )
        self.db_run_query(grant_user_command, grant_user_exception)

    def add_user(self, db_user: str, db_pass: str, db_user_host: str = '%', force: bool = False, timeout=25):
        if self.check_user_exists(db_user, db_user_host):
            if force:
                self.remove_user(db_user, db_user_host)
            else:
                raise DatabaseServiceException(
                    self.run_on_compose_service, f'User {db_user} for {db_user_host} already exists.'
                )

        add_user_command = f"'CREATE USER `{db_user}`@`%` IDENTIFIED BY \"{db_pass}\";'"
        add_user_exception = DatabaseServiceException(self.database_server_info.host, f'Failed to add user {db_user}.')
        self.db_run_query(add_user_command, add_user_exception)

    def db_export(self, db_name: str, export_file_path: Union[str, Path]):
        if not self.check_db_exists(db_name):
            raise DatabaseServiceDBNotFoundError(db_name, self.run_on_compose_service)

        if isinstance(export_file_path, Path):
            export_file_path = str(export_file_path.absolute())

        db_export_command = f"mysqldump -u'{self.database_server_info.user}' -p'{self.database_server_info.password}' -h'{self.database_server_info.host}' -P{self.database_server_info.port} {db_name} --result-file={export_file_path}"

        try:
            output = self.compose_project.docker.compose.exec(
                self.run_on_compose_service, command=db_export_command, stream=False
            )
        except DockerException:
            raise DatabaseServiceDBExportFailed(self.run_on_compose_service, db_name)

    def db_create(self, db_name):
        create_db_command = f"'CREATE DATABASE IF NOT EXISTS `{db_name}`';"
        create_db_exception = DatabaseServiceDBCreateFailed(self.run_on_compose_service, db_name)
        self.db_run_query(create_db_command, create_db_exception)

    def db_import(self, db_name: str, host_db_file_path: Path, force: bool = False):
        if not self.check_db_exists(db_name):
            if force:
                self.db_create(db_name)
            else:
                raise DatabaseServiceDBNotFoundError(db_name, self.run_on_compose_service)

        container_db_file_name = host_db_file_path.name
        source = str(host_db_file_path.absolute())
        destination = f"{self.run_on_compose_service}:/tmp/{container_db_file_name}"
        db_import_command = self.base_command + f" {db_name} -e 'source /tmp/{container_db_file_name}'"
        try:
            output = self.compose_project.docker.compose.cp(source, destination, stream=False)
            output = self.compose_project.docker.compose.exec(
                self.run_on_compose_service, command=db_import_command, stream=False
            )
        except DockerException:
            raise DatabaseServiceDBImportFailed(self.run_on_compose_service, source)



import os
import platform
import shutil
from datetime import datetime
from copy import deepcopy
from pathlib import Path
from typing import Any
from frappe_manager import CLI_DIR
from frappe_manager.compose_manager.ComposeFile import ComposeFile
from frappe_manager.compose_project.compose_project import ComposeProject
from frappe_manager.docker_wrapper.DockerException import DockerException
from frappe_manager.migration_manager.backup_manager import BackupData
from frappe_manager.migration_manager.migration_base import MigrationBase
from frappe_manager.migration_manager.migration_exections import MigrationExceptionInBench
from frappe_manager.migration_manager.migration_helpers import (
    MigrationBench,
    MigrationBenches,
    MigrationServicesManager,
)
from frappe_manager.services_manager.database_service_manager import (
    DatabaseServerServiceInfo,
    DatabaseServiceManager,
    MariaDBManager,
)
from frappe_manager.services_manager.services_exceptions import (
    DatabaseServiceException,
    ServicesException,
    ServicesNotCreated,
)
from frappe_manager.site_manager.site_exceptions import BenchDockerComposeFileNotFound
from frappe_manager.utils.docker import host_run_cp
from frappe_manager.display_manager.DisplayManager import richprint
from frappe_manager.utils.helpers import get_container_name_prefix, get_unix_groups, random_password_generate
from frappe_manager.migration_manager.backup_manager import BackupManager
from frappe_manager.migration_manager.version import Version


class MigrationV0100(MigrationBase):
    version = Version("0.10.0")

    def init(self):
        self.benches_dir = CLI_DIR / "sites"
        self.backup_manager = BackupManager(name=str(self.version), benches_dir=self.benches_dir)
        self.string_timestamp = datetime.now().strftime("%d-%b-%y--%H-%M-%S")
        self.benches_manager = MigrationBenches(self.benches_dir)
        self.services_manager: MigrationServicesManager = MigrationServicesManager()

    def get_rollback_version(self):
        # this was without any migrations
        return Version("0.10.1")

    def up(self):
        richprint.stdout.rule(f':package: [bold][blue]v{str(self.version)}[/blue][bold]')
        self.logger.info(f"v{str(self.version)}: Started")
        self.logger.info("-" * 40)

        self.init()
        self.migrate_services()
        self.migrate_benches()

        self.logger.info("-" * 40)

    def migrate_services(self):
        # create services
        self.services_create(self.services_manager.compose_project)
        self.services_manager.compose_project.pull_images()
        self.services_manager.compose_project.start_service(force_recreate=True)

        self.services_database_manager: DatabaseServiceManager = MariaDBManager(
            DatabaseServerServiceInfo.import_from_compose_file('global-db', self.services_manager.compose_project),
            self.services_manager.compose_project,
        )

    def migrate_benches(self):
        main_error = False

        # migrate each bench
        for bench_name, bench_path in self.benches_manager.get_all_benches().items():
            bench = MigrationBench(name=bench_name, path=bench_path.parent)

            if bench.name in self.migration_executor.migrate_benches.keys():
                bench_info = self.migration_executor.migrate_benches[bench.name]
                if bench_info['exception']:
                    richprint.print(f"Skipping migration for failed bench {bench.name}.")
                    main_error = True
                    continue

            self.migration_executor.set_bench_data(bench, migration_version=self.version)
            try:
                self.migrate_bench(bench)
            except Exception as e:
                import traceback

                traceback_str = traceback.format_exc()
                self.logger.error(f"{bench.name} [ EXCEPTION TRACEBACK ]:\n {traceback_str}")
                richprint.update_live()
                main_error = True
                self.migration_executor.set_bench_data(bench, e, self.version)

                # restore all backup files
                for backup in self.backup_manager.backups:
                    if backup.bench == bench.name:
                        self.backup_manager.restore(backup, force=True)

                self.undo_bench_migrate(bench)
                self.logger.info(f'Undo successfull for bench: {bench.name}')
                bench.compose_project.down_service(volumes=False, timeout=5)

        if main_error:
            raise MigrationExceptionInBench('')

    def migrate_bench(self, bench: MigrationBench):
        richprint.change_head("Migrating bench compose")

        bench.compose_project.down_service(volumes=False)

        self.migrate_bench_compose(bench)

        bench.compose_project.down_service(volumes=False)

        bench_db_info = bench.get_db_connection_info()
        bench_db_name = bench_db_info["name"]
        bench_db_user = bench_db_info["user"]
        bench_db_pass = bench_db_info["password"]

        self.services_database_manager.add_user(bench_db_user, bench_db_pass, force=True)
        self.services_database_manager.grant_user_privilages(bench_db_name, bench_db_user)

    def undo_services_migrate(self):
        self.services_manager.compose_project.down_service()
        if self.services_manager.services_path.exists():
            shutil.rmtree(self.services_manager.services_path)

    def undo_bench_migrate(self, bench: MigrationBench):
        configs_backup = bench.path / f"configs-{self.string_timestamp}.bak"
        configs_path = bench.path / "configs"

        if configs_path.exists():
            shutil.rmtree(configs_path)

        if configs_backup.exists():
            shutil.copytree(configs_backup, configs_path)
            shutil.rmtree(configs_backup)
            self.logger.info(f"Removed : {configs_backup}")

        # start each bench forcefully and add user to the bench
        try:
            bench.compose_project.start_service(['mariadb'])
            bench_db_info = DatabaseServerServiceInfo.import_from_compose_file('mariadb', bench.compose_project)
            bench_db_manager = MariaDBManager(bench_db_info, bench.compose_project)
            bench_db_info = bench.get_db_connection_info()
            bench_db_name = bench_db_info["name"]
            bench_db_user = bench_db_info["user"]
            bench_db_pass = bench_db_info["password"]
            try:
                bench_db_manager.add_user(db_user=bench_db_user, db_pass=bench_db_pass, force=True)
                bench_db_manager.grant_user_privilages(bench_db_user, bench_db_name)
            except DatabaseServiceException as e:
                pass
            bench.compose_project.down_service(volumes=False)
        except DockerException as e:
            pass
        self.logger.info(f"Undo successfull for bench: {bench.name}")

    def migrate_bench_compose(self, bench: MigrationBench):
        richprint.change_head("Migrating database")

        if not bench.compose_project.compose_file_manager.exists():
            raise BenchDockerComposeFileNotFound(bench.name, bench.compose_project.compose_file_manager.compose_path)

        db_backup_file = self.db_migration_export(bench)

        richprint.print(f"[blue]{bench.name}[/blue] db exported from bench mariadb service.")

        # backup bench db
        db_backup = self.backup_manager.backup(db_backup_file, bench_name=bench.name, allow_restore=False)

        self.db_migration_import(bench=bench, db_backup_file=db_backup)

        richprint.print(f"[blue]{bench.name}[/blue] db imported to global-db service.")

        richprint.change_head("Migrating bench compose")

        # get all the payloads
        envs = bench.compose_project.compose_file_manager.get_all_envs()
        labels = bench.compose_project.compose_file_manager.get_all_labels()

        # introduced in v0.10.0
        if not "ENVIRONMENT" in envs["frappe"]:
            envs["frappe"]["ENVIRONMENT"] = "dev"

        envs["frappe"]["CONTAINER_NAME_PREFIX"] = get_container_name_prefix(bench.name)
        envs["frappe"]["MARIADB_ROOT_PASS"] = "root"
        envs["frappe"]["MARIADB_HOST"] = "global-db"

        envs["nginx"]["VIRTUAL_HOST"] = bench.name

        envs["adminer"] = {"ADMINER_DEFAULT_SERVER": "global-db"}

        envs_user_info = {}
        userid_groupid: dict = {"USERID": os.getuid(), "USERGROUP": os.getgid()}

        env_user_info_container_list = ["frappe", "schedule", "socketio"]

        for env in env_user_info_container_list:
            envs_user_info[env] = deepcopy(userid_groupid)

        # overwrite user for each invocation
        users = {"nginx": {"uid": os.getuid(), "gid": os.getgid()}}

        self.create_compose_dirs(bench)

        bench.compose_project.compose_file_manager.template_name = "docker-compose.tmpl"

        # load template
        bench.compose_project.compose_file_manager.yml = bench.compose_project.compose_file_manager.load_template()

        # set all the payload
        bench.compose_project.compose_file_manager.set_all_envs(envs)
        bench.compose_project.compose_file_manager.set_all_envs(envs_user_info)
        bench.compose_project.compose_file_manager.set_all_labels(labels)
        bench.compose_project.compose_file_manager.set_all_users(users)

        bench.compose_project.compose_file_manager.remove_secrets_from_container("frappe")
        bench.compose_project.compose_file_manager.remove_root_secrets_compose()
        bench.compose_project.compose_file_manager.set_network_alias("nginx", "site-network", [bench.name])
        bench.compose_project.compose_file_manager.set_container_names(get_container_name_prefix(bench.name))

        bench.compose_project.compose_file_manager.set_version(str(self.version))
        bench.compose_project.compose_file_manager.set_top_networks_name(
            "site-network", get_container_name_prefix(bench.name)
        )
        bench.compose_project.compose_file_manager.write_to_file()

        # change the node socketio port
        bench.common_bench_config_set({"socketio_port": "80"})

        richprint.print(f"Migrated [blue]{bench.name}[/blue] compose file.")

        return db_backup

    def create_compose_dirs(self, bench: MigrationBench):
        richprint.change_head("Creating services compose dirs.")

        # directory creation
        configs_path = bench.path / "configs"

        # custom config directory found moving it
        # check if config directory exits if exists then move it
        if configs_path.exists():
            backup_path = f"{configs_path.absolute()}.{self.string_timestamp}.bak"
            shutil.move(configs_path, backup_path)

        configs_path.mkdir(parents=True, exist_ok=True)

        # create nginx dirs
        nginx_dir = configs_path / "nginx"
        nginx_dir.mkdir(parents=True, exist_ok=True)

        nginx_poluate_dir = ["conf"]

        nginx_image = bench.compose_project.compose_file_manager.yml["services"]["nginx"]["image"]

        for directory in nginx_poluate_dir:
            new_dir = nginx_dir / directory
            if not new_dir.exists():
                new_dir_abs = str(new_dir.absolute())
                host_run_cp(
                    nginx_image,
                    source="/etc/nginx",
                    destination=new_dir_abs,
                    docker=bench.compose_project.docker,
                )

        nginx_subdirs = ["logs", "cache", "run"]

        for directory in nginx_subdirs:
            new_dir = nginx_dir / directory
            new_dir.mkdir(parents=True, exist_ok=True)

        richprint.print("Created services compose dirs.")

    def db_migration_export(self, bench: MigrationBench) -> Path:
        self.logger.debug("[db export] bench: %s", bench.name)

        # start the benc hand also handle the missing docker images
        output = bench.compose_project.docker.compose.up(
            services=["mariadb", "frappe"], detach=True, pull="missing", stream=False
        )

        self.logger.debug("[db export] checking if mariadb started")

        bench_db_server_info = DatabaseServerServiceInfo.import_from_compose_file('mariadb', bench.compose_project)
        bench_mariadb_manager = MariaDBManager(bench_db_server_info, bench.compose_project)

        # wait till db started
        bench_mariadb_manager.wait_till_db_start()

        # create dir to store migration
        db_migration_dir_path = bench.path / "workspace" / "migrations"
        db_migration_dir_path.mkdir(exist_ok=True)

        current_datetime = datetime.now()
        formatted_date = current_datetime.strftime("%d-%m-%Y--%H-%M-%S")

        db_migration_file_path = f"/workspace/migrations/db-{bench.name}-{formatted_date}.sql"

        bench_db_info = bench.get_db_connection_info()
        bench_db_name = bench_db_info["name"]

        bench_frappe_db_manager = MariaDBManager(bench_db_server_info, bench.compose_project, 'frappe')
        bench_frappe_db_manager.db_export(bench_db_name, db_migration_file_path)

        output_stop = bench.compose_project.docker.compose.stop(timeout=10, stream=False)

        bench_db_migration_file_path = Path(bench.path / db_migration_file_path[1:])

        return bench_db_migration_file_path

    def db_migration_import(self, bench: MigrationBench, db_backup_file: BackupData):
        self.logger.info(f"[database import: global-db] {bench.name} -> {db_backup_file}")

        bench_db_info = bench.get_db_connection_info()
        bench_db_name = bench_db_info["name"]
        bench_db_user = bench_db_info["user"]
        bench_db_pass = bench_db_info["password"]

        # wait till db starts
        self.services_database_manager.wait_till_db_start()

        # import db if db found then remove it and add
        self.services_database_manager.db_import(bench_db_name, db_backup_file.src, True)

        # add user if exits then remove and add
        self.services_database_manager.add_user(bench_db_user, bench_db_pass, force=True)

        self.services_database_manager.grant_user_privilages(bench_db_user, bench_db_name)

        richprint.print(f"{bench.name} db imported.")

    def services_create(self, services_compose_project: ComposeProject):
        richprint.change_head("Creating services.")

        envs = {
            "global-db": {
                "MYSQL_ROOT_PASSWORD_FILE": '/run/secrets/db_root_password',
                "MYSQL_DATABASE": "root",
                "MYSQL_USER": "admin",
                "MYSQL_PASSWORD_FILE": '/run/secrets/db_password',
            }
        }
        current_system = platform.system()
        inputs: dict[str, Any] = {"environment": envs}
        try:
            user = {
                "global-db": {
                    "uid": os.getuid(),
                    "gid": os.getgid(),
                }
            }

            if not current_system == "Darwin":
                user["global-nginx-proxy"] = {
                    "uid": os.getuid(),
                    "gid": get_unix_groups()["docker"],
                }
            inputs["user"] = user

        except KeyError:
            raise ServicesException(
                "docker group not found in system. Please add docker group to the system and current user to the docker group."
            )

        if self.services_manager.services_path.exists():
            shutil.rmtree(self.services_manager.services_path)

        # create required directories
        dirs_to_create = [
            "mariadb/conf",
            "mariadb/logs",
            "nginx-proxy/dhparam",
            "nginx-proxy/certs",
            "nginx-proxy/confd",
            "nginx-proxy/htpasswd",
            "nginx-proxy/vhostd",
            "nginx-proxy/html",
            "nginx-proxy/logs",
            "nginx-proxy/run",
            "nginx-proxy/cache",
            "secrets",
        ]

        # set secrets in compose
        self.generate_compose(inputs, services_compose_project.compose_file_manager)

        if current_system == "Darwin":
            services_compose_project.compose_file_manager.remove_container_user('global-nginx-proxy')
            services_compose_project.compose_file_manager.remove_container_user('global-db')
        else:
            dirs_to_create.append("mariadb/data")

        # create dirs
        for folder in dirs_to_create:
            temp_dir = self.services_manager.services_path / folder
            try:
                temp_dir.mkdir(parents=True, exist_ok=True)
            except Exception as e:
                richprint.exit(f"Failed to create global services bind mount directories. Error: {e}")

        # populate secrets for db
        db_password_path = self.services_manager.services_path / 'secrets' / 'db_password.txt'
        db_root_password_path = self.services_manager.services_path / 'secrets' / 'db_root_password.txt'

        db_password_path.write_text(random_password_generate(password_length=16, symbols=True))
        db_root_password_path.write_text(random_password_generate(password_length=24, symbols=True))

        # populate mariadb config
        mariadb_conf = self.services_manager.services_path / "mariadb/conf"
        mariadb_conf = str(mariadb_conf.absolute())
        host_run_cp(
            image="mariadb:10.6",
            source="/etc/mysql/.",
            destination=mariadb_conf,
            docker=services_compose_project.docker,
        )

        services_compose_project.compose_file_manager.set_secret_file_path(
            'db_password', str(db_password_path.absolute())
        )
        services_compose_project.compose_file_manager.set_secret_file_path(
            'db_root_password', str(db_root_password_path.absolute())
        )
        services_compose_project.compose_file_manager.write_to_file()

        services_compose_project.docker.compose.down(remove_orphans=True, timeout=1, volumes=True, stream=True)

        richprint.print(f"Created services at {self.services_manager.services_path}.")

    def generate_compose(self, inputs: dict, compose_file_manager: ComposeFile):
        richprint.change_head(f"Generating services compose file.")
        try:
            # handle environment
            if "environment" in inputs.keys():
                environments: dict = inputs["environment"]
                compose_file_manager.set_all_envs(environments)

            # handle lablels
            if "labels" in inputs.keys():
                labels: dict = inputs["labels"]
                compose_file_manager.set_all_labels(labels)

            # handle user
            if "user" in inputs.keys():
                user: dict = inputs["user"]
                for container_name in user.keys():
                    uid = user[container_name]["uid"]
                    gid = user[container_name]["gid"]
                    compose_file_manager.set_user(container_name, uid, gid)
            richprint.print(f"Generated services compose file.")
        except Exception:
            raise ServicesNotCreated()



from builtins import len
from pathlib import Path
from typing import List, Optional

from frappe_manager.docker_wrapper.subprocess_output import SubprocessOutput
from frappe_manager.utils import helpers
from rich.box import Box
from rich.style import Style
from typer import Option


class BenchException(Exception):
    def __init__(
        self,
        bench_name: str,
        message: str,
        prefix_bench_name: bool = True,
    ):
        self.message = message

        if prefix_bench_name:
            self.message = f"[blue][bold]{bench_name} :[/bold][/blue] {message}"

        super().__init__(self.message)


class BenchDockerComposeFileNotFound(BenchException):
    def __init__(
        self,
        bench_name: str,
        path: Path,
        message: str = "Compose file not found at {}. Aborting operation.",
    ):
        self.bench_name = bench_name
        self.path = path
        self.message = message.format(self.path)
        super().__init__(self.bench_name, self.message)


class BenchServiceNotRunning(BenchException):
    def __init__(
        self,
        bench_name: str,
        service: str,
        message: str = "Service {} not running.",
    ):
        self.bench_name = bench_name
        self.service = service
        self.message = message.format(self.service)
        super().__init__(self.bench_name, self.message)


class BenchNotFoundError(BenchException):
    def __init__(
        self,
        bench_name: str,
        path: Path,
        message: str = "Bench not found at {}.",
    ):
        self.bench_name = bench_name
        self.path = path
        self.message = message.format(self.path)
        super().__init__(self.bench_name, self.message)


class BenchRemoveDirectoryError(BenchException):
    def __init__(
        self,
        bench_name: str,
        path: Path,
        message: str = "Remove dirs failed at {}.",
    ):
        self.bench_name = bench_name
        self.path = path
        self.message = message.format(self.path)
        super().__init__(self.bench_name, self.message)


class BenchLogFileNotFoundError(BenchException):
    def __init__(
        self,
        bench_name: str,
        path: Path,
        message: str = "Log file not found at {}.",
    ):
        self.bench_name = bench_name
        self.path = path
        self.message = message.format(self.path)
        super().__init__(self.bench_name, self.message)


class BenchWorkersStartError(BenchException):
    def __init__(
        self,
        bench_name: str,
        message: str = "Workers not able to start.",
    ):
        self.bench_name = bench_name
        self.message = message
        super().__init__(self.bench_name, self.message)


class BenchWorkersSupervisorConfigurtionGenerateError(BenchException):
    def __init__(
        self,
        bench_name: str,
        message: str = "Failed to configure workers.",
    ):
        self.bench_name = bench_name
        self.message = message
        super().__init__(self.bench_name, self.message)


class BenchWorkersSupervisorConfigurtionNotFoundError(BenchException):
    def __init__(
        self,
        bench_name: str,
        config_dir: str,
        message: str = "Superviosrd workers configuration not found in {}.",
    ):
        self.bench_name = bench_name
        self.config_dir = config_dir
        self.message = message.format(self.config_dir)
        super().__init__(self.bench_name, self.message)


class BenchConfigFileNotFound(BenchException):
    def __init__(self, bench_name, config_path, message="Config file not found at {}."):
        self.bench_name = bench_name
        self.config_path = config_path
        self.message = message.format(config_path)
        super().__init__(self.bench_name, self.message)


class BenchConfigValidationError(BenchException):
    def __init__(self, bench_name, config_path, message="FM bench config not valid at {}"):
        self.bench_name = bench_name
        self.conig_path = config_path
        self.message = message.format(self.conig_path)
        super().__init__(self.bench_name, self.message)


class AdminToolsFailedToStart(BenchException):
    def __init__(self, bench_name, message="Failed to start admin tools."):
        self.bench_name = bench_name
        self.message = message
        super().__init__(self.bench_name, self.message)


class BenchSSLCertificateAlreadyIssued(BenchException):
    def __init__(self, bench_name, message="SSL Certificate already issued."):
        self.bench_name = bench_name
        self.message = message
        super().__init__(self.bench_name, self.message)


class BenchSSLCertificateNotIssued(BenchException):
    def __init__(self, bench_name, message="No SSL Certificate issued."):
        self.bench_name = bench_name
        self.message = message
        super().__init__(self.bench_name, self.message)


class BenchAttachTocontainerFailed(BenchException):
    def __init__(self, bench_name, service_name, message="Attach to {} service container failed."):
        self.bench_name = bench_name
        self.service_name = service_name
        self.message = message.format(self.service_name)
        super().__init__(self.bench_name, self.message)


class BenchNotRunning(BenchException):
    def __init__(self, bench_name, message="Bench services not running."):
        self.bench_name = bench_name
        self.message = message
        super().__init__(self.bench_name, self.message)


class BenchFailedToRemoveDevPackages(BenchException):
    def __init__(self, bench_name, message="Not able pip uninstall dev packages."):
        self.bench_name = bench_name
        self.message = message
        super().__init__(self.bench_name, self.message)


class BenchFrappeServiceSupervisorNotRunning(BenchException):
    def __init__(self, bench_name, message="Supervisorctl is not running in frappe service"):
        self.bench_name = bench_name
        self.message = message
        super().__init__(self.bench_name, self.message)


class BenchOperationException(BenchException):
    def __init__(
        self,
        bench_name,
        message: str,
        print_combined: bool = True,
        print_stdout: bool = False,
        print_stderr: bool = False,
    ):
        self.bench_name = bench_name
        self.message = message
        self.print_stdout = print_stdout
        self.print_stderr = print_stderr
        self.print_combined = print_combined
        self.output = None
        super().__init__(self.bench_name, self.message)

    def set_output(self, output: SubprocessOutput):
        self.output = output
        from rich.panel import Panel

        to_print = []

        box: Box = Box("╭   \n" "    \n" " ── \n" "│   \n" "    \n" "    \n" " |  \n" "    \n", ascii=True)

        if self.print_stdout:
            panel = Panel.fit(
                "\n".join(self.output.stdout),
                box=box,
                padding=(0, 1),
                border_style="dim",
                title="Error command stdout",
                title_align="left",
            )
            to_print.append(helpers.rich_object_to_string(panel))

        if self.print_combined:
            panel = Panel.fit(
                "\n".join(self.output.combined),
                box=box,
                padding=(0, 1),
                border_style="dim",
                title="Error command output",
                title_align="left",
            )
            to_print.append(helpers.rich_object_to_string(panel))

        if self.print_stderr:
            panel = Panel.fit(
                "\n".join(self.output.stderr),
                box=box,
                padding=(0, 1),
                border_style="dim",
                title="Error command stderr",
                title_align="left",
            )
            to_print.append(helpers.rich_object_to_string(panel))

        self.message = self.message + "\n" + "\n".join(to_print)

        super().__init__(self.bench_name, self.message, prefix_bench_name=False)


class BenchOperationFrappeBranchChangeFailed(BenchException):
    def __init__(self, bench_name, app: str, branch: str, message: str = "Failed to change {} app branch to {}."):
        self.bench_name = bench_name
        self.app = app
        self.branch = branch
        self.message = message.format(app, branch)
        super().__init__(self.bench_name, self.message)


class BenchOperationRequiredDockerImagesNotAvailable(BenchException):
    def __init__(
        self,
        bench_name,
        pull_command,
        message: str = "Required docker images not available. Pull all required images using command '{}'.",
    ):
        self.bench_name = bench_name
        self.message = message.format(pull_command)
        super().__init__(self.bench_name, self.message)


class BenchOperationWaitForRequiredServiceFailed(BenchOperationException):
    def __init__(
        self,
        bench_name,
        host: str,
        port: str,
        timeout: int,
        message: str = "Waiting for service {}:{} timed out. {}",
        print_combined: bool = True,
        print_stdout: bool = False,
        print_stderr: bool = False,
    ):
        self.bench_name = bench_name
        self.host = host
        self.port = port
        self.timeout = timeout
        self.print_stdout = print_stdout
        self.print_stderr = print_stderr
        self.print_combined = print_combined
        self.message = message.format(host, port, timeout)

        super().__init__(self.bench_name, self.message, self.print_combined, self.print_stdout, self.print_stderr)


class BenchOperationBenchSiteCreateFailed(BenchOperationException):
    def __init__(
        self,
        bench_name,
        print_combined: bool = True,
        print_stdout: bool = False,
        print_stderr: bool = False,
        message: str = "Failed to create site {}.",
    ):
        self.bench_name = bench_name
        self.message = message.format(bench_name)
        self.print_stdout = print_stdout
        self.print_stderr = print_stderr
        self.print_combined = print_combined
        super().__init__(self.bench_name, self.message, self.print_combined, self.print_stdout, self.print_stderr)


class BenchOperationBenchInstallAppInPythonEnvFailed(BenchOperationException):
    def __init__(
        self,
        bench_name,
        app_name: str,
        message: str = "Failed to install app {} in python env.",
        print_combined: bool = True,
        print_stdout: bool = False,
        print_stderr: bool = False,
    ):
        self.bench_name = bench_name
        self.app_name = app_name
        self.message = message.format(app_name)
        self.print_stdout = print_stdout
        self.print_stderr = print_stderr
        self.print_combined = print_combined

        super().__init__(self.bench_name, self.message, self.print_combined, self.print_stdout, self.print_stderr)


class BenchOperationBenchRemoveAppFromPythonEnvFailed(BenchOperationException):
    def __init__(
        self,
        bench_name,
        app_name: str,
        message: str = "Failed to remove app {} from python env.",
        print_combined: bool = True,
        print_stdout: bool = False,
        print_stderr: bool = False,
    ):
        self.bench_name = bench_name
        self.app_name = app_name
        self.message = message.format(app_name)
        self.print_stdout = print_stdout
        self.print_stderr = print_stderr
        self.print_combined = print_combined

        super().__init__(self.bench_name, self.message, self.print_combined, self.print_stdout, self.print_stderr)


class BenchOperationBenchAppInSiteFailed(BenchOperationException):
    def __init__(
        self,
        bench_name,
        app_name: str,
        message: str = "Failed to install app {} in site {}.",
        print_combined: bool = True,
        print_stdout: bool = False,
        print_stderr: bool = False,
    ):
        self.bench_name = bench_name
        self.app_name = app_name
        self.message = message.format(app_name, self.bench_name)
        self.print_stdout = print_stdout
        self.print_stderr = print_stderr
        self.print_combined = print_combined
        super().__init__(self.bench_name, self.message, self.print_combined, self.print_stdout, self.print_stderr)


class BenchOperationBenchBuildFailed(BenchOperationException):
    def __init__(
        self,
        bench_name,
        apps: Optional[List[str]] = None,
        message: str = "Failed to build",
        print_combined: bool = True,
        print_stdout: bool = False,
        print_stderr: bool = False,
    ):
        self.bench_name = bench_name
        self.apps = apps
        if apps:
            message = message + " app"
            if len(apps) > 1:
                message = message + " apps"
            for app in apps:
                message += f" {app}"
        self.message = message
        self.print_stdout = print_stdout
        self.print_stderr = print_stderr
        self.print_combined = print_combined
        super().__init__(self.bench_name, self.message, self.print_combined, self.print_stdout, self.print_stderr)



import typer
from typing import Annotated, Optional
from frappe_manager.services_manager.services import ServicesManager
from frappe_manager.display_manager.DisplayManager import richprint
from frappe_manager.services_manager import ServicesEnum

services_root_command = typer.Typer(no_args_is_help=True, rich_markup_mode="rich")

@services_root_command.command(no_args_is_help=True)
def stop(
    ctx: typer.Context,
    service_name: Annotated[ServicesEnum, typer.Argument(help="Name of the service.")],
):
    """Stops global services."""
    services_manager: ServicesManager = ctx.obj["services"]
    if service_name.value == ServicesEnum.all:
        for service in ServicesEnum:
            if service == ServicesEnum.all:
                continue

            if not services_manager.compose_project.is_service_running(service.value):
                richprint.print(f"Skipping not running service {service.value}.")
                continue

            services_manager.compose_project.stop_service(services=[service.value])
            richprint.print(f"Stopped service {service.value}.")
    else:
        if services_manager.compose_project.is_service_running(service_name.value):
            services_manager.compose_project.stop_service(services=[service_name.value])
        else:
            richprint.print(f"Skipping already stopped service {service_name.value}.")


@services_root_command.command(no_args_is_help=True)
def start(
    ctx: typer.Context,
    service_name: Annotated[ServicesEnum, typer.Argument(help="Name of the service.")],
):
    """Starts global services."""
    services_manager: ServicesManager = ctx.obj["services"]

    if service_name.value == ServicesEnum.all:

        for service in ServicesEnum:
            if service == ServicesEnum.all:
                continue

            if services_manager.compose_project.is_service_running(service.value):
                richprint.print(f"Skipping already running service {service.value}.")
                continue

            services_manager.compose_project.start_service(services=[service.value])
            richprint.print(f"Started service {service.value}.")
    else:
        if not services_manager.compose_project.is_service_running(service_name.value):
            services_manager.compose_project.start_service(services=[service_name.value])
        else:
            richprint.print(f"Skipping already running service {service_name.value}.")

@services_root_command.command(no_args_is_help=True)
def restart(
    ctx: typer.Context,
    service_name: Annotated[ServicesEnum, typer.Argument(help="Name of the service.")],
):
    """Restarts global services."""
    services_manager: ServicesManager = ctx.obj["services"]

    if service_name.value == ServicesEnum.all:

        for service in ServicesEnum:
            if service == ServicesEnum.all:
                continue

            services_manager.compose_project.restart_service(services=[service.value])
            richprint.print(f"Restarted service {service.value}.")
    else:
        services_manager.compose_project.restart_service(services=[service_name.value])
        richprint.print(f"Restarted service {service_name.value}.")


@services_root_command.command(no_args_is_help=True)
def shell(
    ctx: typer.Context,
    service_name: Annotated[ServicesEnum, typer.Argument(help="Name of the service.")],
    user: Annotated[Optional[str], typer.Option(help="Connect as this user.")] = None,
):
    """
    Open shell for the specificed global service.
    """
    services_manager: ServicesManager = ctx.obj["services"]
    services_manager.shell(service_name.value, user)
